#include "Correspondence.h"

using namespace std;
using namespace icp2;
using namespace Preprocessing;

namespace Correspondence {

double **distance1 = NULL;
double **distance2 = NULL;
MyMesh::VHandle **voisins1 = NULL;
MyMesh::VHandle **voisins2 = NULL;
double diagonale;
int nb_fitting_levels = 8;
double **tab_feature_vector_s;
double **tab_feature_vector_m;

void Correspondence_Computation(MyMesh *source, MyMesh *model, int nb_source, int nb_model, MyMesh::VertexHandle *** correspondence, double c0, double** poids) {
    list<assignment> kernel;
    int *has_a_correspondence;
    Initial_Correspondence(source, model, nb_source, correspondence);
    Prunning(nb_source, correspondence, c0, &kernel, &has_a_correspondence);
    Propagation(nb_source, &kernel, &has_a_correspondence, correspondence);
    Poids(*correspondence, &kernel, nb_source, poids);
}

void Initial_Correspondence(MyMesh * source, MyMesh * model, int nb_source_subset, MyMesh::VertexHandle *** correspondence) {
    printf("Computing initial correspondence \n");
    Closest_Points (source, model, nb_source_subset, correspondence);
    // On choisit 5% de la diagonale de la bounding box du maillage source comme seuil
    int threshold = 0.05 * diagonale;
    Improve_Correspondence(source, model, (*source).n_vertices(), voisins2, threshold, correspondence, 0);
    //Improve_Correspondence(source,model,(*source).n_vertices(),voisins1,threshold,correspondence,1);
    printf("Initial correspondence computed\n");
}

void Prunning(int nb_vertices, MyMesh::VertexHandle *** correspondence, double c0, list<assignment> *kernel, int **has_a_correspondence) {
    /*
     * Spectral matching proceeds (LH05)
     *
     */
    printf("Computing correspondence prunning \n");
    // 1. Construction de la matrice non negative et symetrique m
    gsl_matrix *m;
    Consistent_Matrix(nb_vertices, *correspondence, &m, c0);
    /* Initialisation d'un tableau de nb_vertices entier (utilise pour la phase de propagation)
     * une valeur 0 a l'indice i signifie que l'element si n'a pas de correspondence
     * une valeur 1 a l'indice i indique que l'element si a un correspondant
     */
    (*has_a_correspondence) = (int *) malloc( nb_vertices * sizeof(int) );

    for (int i = 0; i < nb_vertices; i++) {
        (*has_a_correspondence)[i] = 0;
    }

    // 2. Calcul du vecteur propre principal x* (associe a la valeur propre max)
    gsl_vector *x;
    Eigen_V(m, &x, nb_vertices);
    // Initialisation de l'argument max a* du vecteur propre
    float max, cour;
    double cij;
    int ind_max = 0, ind, nb_elements_traites = 0;
    assignment assignment_cour;

    /*
     * On ajoute iterativement la correspondance qui a le plus haut
     * score dans le noyau seulement si sa mesure de consistence par
     * rapport au autre éléments du noyau est > c0
     *
     */
    do {
        // 2. Calcul de l'argument max a* du vecteur propre
        max = gsl_vector_get(x, 0);
        ind = 0;
        ind_max = 0;

        while (max == -1) {
            ind++;
            max = gsl_vector_get(x, ind);
        }

        ind_max = ind;

        for (int i = ind + 1; i < nb_vertices; i++) {
            // On regarde si l'element n'as deja ete utilise
            cour = gsl_vector_get(x, i);

            if ( cour != -1 && cour > max ) {
                max = cour;
                ind_max = i;
            }
        }

        // On enleve a* de la liste
        gsl_vector_set(x, ind_max, -1);
        assignment_cour.s = (*correspondence)[0][ind_max];
        assignment_cour.t = (*correspondence)[1][ind_max];

        // 3. Si x*(a*) = 0 stop
        if ( max > 0) {
            // Sinon on insere a* dans le noyau seulement si sa mesure de consistence
            // par rapport au autre éléments du noyau est > c0
            bool condition = true;

            for (list<assignment>::const_iterator it = (*kernel).begin(); it != (*kernel).end(); it++) {
                cij = Consistency_Measure(assignment_cour.s, assignment_cour.t, (*it).s, (*it).t);

                if (cij <= c0) {
                    condition = false;
                }
            }

            // Si la condition est verifiee on ajoute l'element dans le noyau
            if (condition) {
                (*kernel).push_back(assignment_cour);
                (*has_a_correspondence)[assignment_cour.s.idx()] = 1;
            }
        }

        nb_elements_traites++;
    } while ( max > 0 && nb_elements_traites < nb_vertices);

    printf("Correspondence prunning computed\n");
}


void Consistent_Matrix(int nb_vertices, MyMesh::VertexHandle ** correspondence, gsl_matrix **m, double c0) {
    double cij;
    *m = gsl_matrix_alloc(nb_vertices, nb_vertices);

    for (int i = 0; i < nb_vertices; i++) {
        for (int j = 0; j < nb_vertices; j++) {
            cij = Consistency_Measure(correspondence[0][i], correspondence[1][i], correspondence[0][j], correspondence[1][j]);

            if ( cij > c0 ) {
                gsl_matrix_set(*m, i, j, (cij - c0)*(cij - c0) / ((1 - c0)*(1 - c0)) );
            } else {
                gsl_matrix_set(*m, i, j, 0);
            }
        }
    }
}


double Consistency_Measure(MyMesh::VertexHandle si, MyMesh::VertexHandle ti, MyMesh::VertexHandle sj, MyMesh::VertexHandle tj) {
    double geodesic_distance1 = 0;
    double geodesic_distance2 = 0;
    double a = distance1[si.idx()][sj.idx()];
    double b = distance2[ti.idx()][tj.idx()];

    if (b != 0) {
        geodesic_distance1 = a / b;
    } else {
        geodesic_distance1 = 0;
    }

    if (a != 0) {
        geodesic_distance2 = b / a;
    } else {
        geodesic_distance2 = 0;
    }

    double min = geodesic_distance1;

    if (geodesic_distance2 < geodesic_distance1) {
        min = geodesic_distance2;
    }

    return min;
}



void Propagation(int nb_vertices, list<assignment> *kernel, int **has_a_correspondence, MyMesh::VertexHandle *** correspondence) {
    printf("Computing correspondence propagation \n");

    for (int i = 0; i < nb_vertices; i++) {
        if ( (*has_a_correspondence)[i] == 0 ) {
            /* Si l'element si n'a pas de correspondant on lui assigne une cible
             * la plus consitante possible par rapport au noyau de correspondances
             * et au critere de distance geodesique
             */
            (*correspondence)[0][i] = MyMesh::VertexHandle(i);
            (*correspondence)[1][i] = Assign(kernel, i);
        }
    }

    /* Ajout des elements du noyau dans le tableau de correspondances */

    for (list<assignment>::const_iterator iit = (*kernel).begin(); iit != (*kernel).end(); iit++) {
        (*correspondence)[0][(*iit).s.idx()] = (*iit).s;
        (*correspondence)[1][(*iit).s.idx()] = (*iit).t;
    }

    printf("Correspondence propagation computed\n");
}

MyMesh::VertexHandle Assign(list<assignment> *kernel, int si) {
    int tj = 0;
    double cour, min;
    /*
     * Calcul du sj de la correspondance (sj,tj) la plus proche de si
     *
     */
    min = numeric_limits<double>::max();

    for (list<assignment>::const_iterator iit = (*kernel).begin(); iit != (*kernel).end(); iit++) {
        cour = distance1[si][(*iit).s.idx()];

        if ( cour < min ) {
            tj = (*iit).t.idx();
            min = cour;
        }
    }

    MyMesh::VertexHandle argmin = MyMesh::VertexHandle(tj);
    min = Consistency_Error(kernel, si, tj);

    for (int j = 0; j < 15; j++) {
        cour = Consistency_Error(kernel, si, voisins2[tj][j].idx());

        if ( cour < min ) {
            min = cour;
            argmin = voisins2[tj][j];
        }
    }

    return argmin;
}

double Consistency_Error(list<assignment> *kernel, int s, int t) {
    double sum = 0;
    double dg1, dg2;

    for (list<assignment>::const_iterator iit = (*kernel).begin(); iit != (*kernel).end(); iit++) {
        dg1 = distance1[s][(*iit).s.idx()];
        dg2 = distance2[t][(*iit).t.idx()];
        sum += (dg1 - dg2) * (dg1 - dg2);
    }

    return sum;
}

void Fine_Tuning(MyMesh *source , MyMesh  * model, MyMesh::VertexHandle ** correspondence) {
}



void Improve_Correspondence(MyMesh *source, MyMesh *model, int nb_source, MyMesh::VertexHandle  ** neighbors, double threshold, MyMesh::VertexHandle *** correspondence, int etape) {
    double norm, cour;
    MyMesh::VertexHandle *tab_voisins;
    MyMesh::VertexHandle i, j, valj;
    MyMesh::Point p1, p2;
    double dist;
    int progress = 0;
    printf("Computing improve correspondence \n");
    printf("[");

    for (int indi = 0; indi < nb_source; indi++) {
        if (indi > (progress + 1) * nb_source / 20) {
            progress = indi * 20 / nb_source;
            printf("#");
        }

        i = (*correspondence)[etape][indi];
        tab_voisins = neighbors[indi];
        valj = (*correspondence)[1-etape][indi];

        if ( etape == 0) {
            norm = Normal(source, model, i, valj);
        } else {
            norm = Normal(source, model, valj, i);
        }

        for (int indj = 0; indj < 15; indj++) {
            j = tab_voisins[indj];

            if (etape == 0) {
                cour = Normal(source, model, i, j);
            } else {
                cour = Normal(source, model, j, i);
            }

            if ( cour < norm ) {
                norm = cour;
                valj = j;
            }
        }

        /*
         * etape = 0:
         *   a chaque element de la source si on modifie son correspondant ti dans
         * le maillage cible en choisissant le meilleur parmis ses voisins
         *
         * etape = 1:
         *   a chaque element de la cible ti on modifie son correspondant dans le
         * maillage source en choisissant le meilleur element si_prime
         * parmis ses voisins
         *
         * On ecarte la correspondance si il ya eu un trop grand mouvement
         */
        if ( etape == 1 ) {
            p1 = Access_Point(source, valj);
            p2 = Access_Point(source, (*correspondence)[0][indi]);
            dist = Norm(&p1, &p2, 3);

            if ( dist < threshold) {
                (*correspondence)[0][indi] = valj;
            }
        } else {
            (*correspondence)[1][indi] = valj;
        }
    }

    printf("]\n");
    printf("Improve correspondence computed\n");
}


double Normal(MyMesh *source, MyMesh *model , MyMesh::VHandle p1,  MyMesh::VHandle p2) {
    double *v1, *v2;
    v1 = tab_feature_vector_s[p1.idx()];
    v2 = tab_feature_vector_m[p2.idx()];
    return Norm(v1, v2, nb_fitting_levels);
}


double Norm(double *v1, double *v2, int n) {
    double res = 0;

    for (int i = 0; i < n; i++) {
        res += (v1[i] - v2[i]) * (v1[i] - v2[i]);
    }

    return sqrt(res);
}

double Norm(MyMesh::Point *v1, MyMesh::Point *v2, int n) {
    double res = 0;

    for (int i = 0; i < n; i++) {
        res += sqrt((float) ((*v1)[i] - (*v2)[i]) * ((*v1)[i] - (*v2)[i]) );
    }

    return res;
}

MyMesh::Point Access_Point(MyMesh *m,  MyMesh::VertexHandle num_vertice) {
    MyMesh::Point p;
    p = (*m).point(num_vertice);
    return p;
}

void Poids(MyMesh::VertexHandle ** correspondence, list<assignment> *kernel, int nb_vertices, double **resultat) {
    double wi;
    // Calcul de la consistence e du noyau de correspondence
    double e = 0;
    int n_kernel = (*kernel).size();
    *resultat = (double*) malloc(nb_vertices * sizeof(double));

    for (list<assignment>::const_iterator iit = (*kernel).begin(); iit != (*kernel).end(); iit++) {
        e += Consistency_Error(kernel, (*iit).s.idx(), (*iit).t.idx());
    }

    if ( e == 0) {
        e = 1;
    }

    e = e / (double)n_kernel;

    // Calcul des poids pour chaque correspondence d'indice i
    for (int i = 0; i < nb_vertices; i++) {
        wi = Consistency_Error(kernel, (correspondence[0][i]).idx() , (correspondence[1][i]).idx() );
        (*resultat)[i] = exp(-wi / (2 * e));
    }
}

void Affect_Preprocessing_Correspondence(double **distance_1,  double **distance_2,  MyMesh::VHandle **voisins_1,  MyMesh::VHandle **voisins_2, double diag, double **tab_feature_vector_source, double **tab_feature_vector_model) {
    distance1 = distance_1;
    distance2 = distance_2;
    voisins1 = voisins_1;
    voisins2 = voisins_2;
    diagonale = diag;
    tab_feature_vector_s = tab_feature_vector_source;
    tab_feature_vector_m = tab_feature_vector_model;
}



}
