#include "Non_Rigid_Registration.h"
#include <iostream>
#include <fstream>

using namespace std;
using namespace icp2;
using namespace Preprocessing;
using namespace Correspondence;
using namespace Clusters;
using namespace energy;

namespace Non_Rigid_Registration {

double **distance_1 = NULL;
double **distance_2 = NULL;
MyMesh::VHandle **voisins1 = NULL;
MyMesh::VHandle **voisins2 = NULL;
int nb_source;
int nb_model;
double diag;

bool save = false;
bool saved = false;

void save_data(double** data, const char* file_name, const int size);
void save_data(MyMesh::VertexHandle **data1, double* data2, const char* file_name, const int size);
void load_data(double*** data, const char* file_name, const int size);
void load_data(MyMesh::VertexHandle ***data1, double** data2, const char* file_name, const int size);

void Compute_Preprocessing(MyMesh *source, MyMesh *model) {
    int nb_fitting_levels = 8;
    /*
     * Calcul de la bounding box
     *
     */
    MyMesh::Point p1, p2;
    BoundingBox(source, &p1, &p2);
    diag = distance2(&p1, &p2);
    distance_1 = NULL;
    distance_2 = NULL;
    /*
    * Calcul du nombre de sommets de la source et du model
    *
    */
    nb_source = (*source).n_vertices();
    nb_model = (*model).n_vertices();

    /*
    * Calcul du tableau des distances geodesique de la source
    */
    if (saved) {
        cout << "Geodesic distances in source : ";
        load_data(&distance_1, "distance_1.sav", source->n_vertices());
    } else {
        G_distance(source, &distance_1);

        if (save) {
            save_data(distance_1, "distance_1.sav", source->n_vertices());
        }
    }

    /*
     *
     * Calcul des ensemble de voisins pour chaque sommets de la source
     * On recopie le tableau des distances dans un temporaire
     * pour ne pas etre ecrase dans la fonction Neighborhoods
     *
     */
    double **temp = (double **) malloc(nb_source * sizeof(double *));

    for (int i = 0; i < nb_source; i++) {
        temp[i] = (double *) malloc( nb_source * sizeof(double) );
    }

    for (int i = 0; i < nb_source; i++) {
        for (int j = 0; j < nb_source; j++) {
            temp[i][j] = distance_1[i][j];
        }
    }

    Neighborhoods(source, &voisins1, temp);
    free(temp);
    /*
     * Calcul du tableau des distances geodesique du model
     */

    if (saved) {
        cout << "Geodesic distances in model : ";
        load_data(&distance_2, "distance_2.sav", model->n_vertices());
    } else {
        G_distance(model, &distance_2);

        if (save) {
            save_data(distance_2, "distance_2.sav", model->n_vertices());
        }
    }

    /*
     *
     * Calcul des ensemble de voisins pour chaque sommets du model
     * On recopie le tableau des distances dans un temporaire
     * pour ne pas etre ecrase dans la fonction Neighborhoods
     *
     */
    double **temp2 = (double **) malloc(nb_model * sizeof(double *));

    for (int i = 0; i < nb_model; i++) {
        temp2[i] = (double *) malloc( nb_model * sizeof(double) );
    }

    for (int i = 0; i < nb_model; i++) {
        for (int j = 0; j < nb_model; j++) {
            temp2[i][j] = distance_2[i][j];
        }
    }

    Neighborhoods(model, &voisins2, temp2);
    free(temp2);
    /*
     *
     * Calcul des features vector
     *
     *
     */
    double **tab_feature_vector_source = (double **) malloc(nb_source * sizeof(double *));

    for (int i = 0; i < nb_source; i++) {
        tab_feature_vector_source[i] = (double *) malloc( nb_fitting_levels * sizeof(double) );
    }

    for (int i = 0; i < nb_source; i++) {
        Compute_Feature_V(source, MyMesh::VertexHandle(i), tab_feature_vector_source[i], distance_1, diag);
    }

    double **tab_feature_vector_model = (double **) malloc(nb_model * sizeof(double *));

    for (int i = 0; i < nb_model; i++) {
        tab_feature_vector_model[i] = (double *) malloc( nb_fitting_levels * sizeof(double) );
    }

    for (int i = 0; i < nb_model; i++) {
        Compute_Feature_V(model, MyMesh::VertexHandle(i), tab_feature_vector_model[i], distance_2, diag);
    }

    /*
     *
     * Charge les elements calcules pour l'etape des correspondences
     *
     */
    Affect_Preprocessing_Correspondence(distance_1, distance_2, voisins1, voisins2, diag, tab_feature_vector_source, tab_feature_vector_model);
    return;
}


void Compute_Non_Rigid_Registration(MyMesh *source_ini, MyMesh *source, MyMesh *model, list<Cluster*>** ls_clust, double c0, double E_max, MyMesh::VertexHandle *** resultat, int N) {
    printf("Compute_Non_Rigid_Registration \n");
    MyMesh::VertexHandle* pts_src;
    MyMesh::VertexHandle* pts_target;
    MyMesh::VertexHandle** neighbors = NULL;
    double **distances = NULL;
    int nb_neighbors = 15;
    double* poids;
    list<Sample*>* s;
    Copie_Mesh_Handle(source, &pts_src, (*source).n_vertices());
    Copie_Mesh_Handle(model, &pts_target, (*model).n_vertices());
    /*
     * Etape de preprocessing
     *
     */
    Compute_Preprocessing(source, model);
    int nb_etape = 0;

    /*
     * Calcul des correspondence
     *
     */
    if (saved) {
        cout << "Correspondence computation : ";
        load_data(resultat, &poids, "correspondence.sav", source->n_vertices());
    } else {
        Correspondence_Computation(source, model, (*source).n_vertices(), (*model).n_vertices(), resultat, c0, &poids);

        if (save) {
            save_data(*resultat, poids, "correspondence.sav", source->n_vertices());
        }
    }

    /*
     * Initialisation de la liste des clusters
     *
     */
    init_clusters(ls_clust, source, &pts_src, model, &pts_target, resultat, &voisins1, (*source).n_vertices(), nb_neighbors);
    /*
     * Calcul des clusters
     *
     */
    process_clustering(*ls_clust, source, &pts_src, model, &pts_target, &voisins1, resultat, nb_neighbors, E_max);
    /*
     * Initialisation de l'etape de minimisation de l'energie
     *
     */
    Init_E(source_ini, source, model, &voisins1, &s, *ls_clust, *resultat, poids);
    /*
     * Etape principale de minimisation de l'energie
     *
     */
    Minimise_E(source_ini, source, model, s, *ls_clust);
    nb_etape++;
    return;
}

void set_save() {
    save = true;
}

void set_saved() {
    saved = true;
}

void save_data(double** data, const char* file_name, const int size) {
    cout << "Save data..." << endl;
    ofstream out(file_name);

    if (!out) {
        cout << "Cannot open file.\n";
    } else {
        for (int i = 0; i < size; ++i) {
            for (int j = 0; j < size; ++j) {
                out << data[i][j] << endl;
            }
        }
    }

    out.close();
}

void save_data(MyMesh::VertexHandle **data1, double* data2, const char* file_name, const int size) {
    cout << "Save data..." << endl;
    ofstream out(file_name);

    if (!out) {
        cout << "Cannot open file.\n";
    } else {
        for (int i = 0; i < size; ++i) {
            out << data2[i] << endl;
            out << (data1[0][i]).idx() << endl;
            out << (data1[1][i]).idx() << endl;
        }
    }

    out.close();
}

void load_data(double*** data, const char* file_name, const int size) {
    cout << "Load data..." << endl;
    ifstream in(file_name);

    if (!in) {
        cout << "Cannot open file.\n";
    } else {
        (*data) = (double **) malloc(size * sizeof(double*));

        for (int i = 0; i < size; ++i) {
            (*data)[i] = (double*) malloc(size * sizeof(double));
        }

        for (int i = 0; i < size; ++i) {
            for (int j = 0; j < size; ++j) {
                in >> (*data)[i][j];
            }
        }
    }

    in.close();
}

void load_data(MyMesh::VertexHandle ***data1, double** data2, const char* file_name, const int size) {
    cout << "Load data..." << endl;
    ifstream in(file_name);

    if (!in) {
        cout << "Cannot open file.\n";
    } else {
        (*data1) = (MyMesh::VertexHandle **) malloc(2 * sizeof(MyMesh::VertexHandle*) );

        for (int i = 0; i < 2; ++i) {
            (*data1)[i] = (MyMesh::VertexHandle* ) malloc(size * sizeof(MyMesh::VertexHandle) );
        }

        *data2 = (double*) malloc(size * sizeof(double));
        int v;

        for (int i = 0; i < size; ++i) {
            in >> (*data2)[i];
            in >> v;
            (*data1)[0][i] = MyMesh::VertexHandle(v);
            in >> v;
            (*data1)[1][i] = MyMesh::VertexHandle(v);
        }
    }

    in.close();
}

}
