#include "Preprocessing.h"
#define min(a,b) (a<=b?a:b)

using namespace std;
namespace Preprocessing {


int nb_neighbors = 15;

int nb_fitting_levels = 4;


void BoundingBox(MyMesh *source, MyMesh::Point *p1, MyMesh::Point *p2) {
    cout << "Computing the bounding box" << endl;
    // Declaration des variables temporaires
    MyMesh::VertexIter v_it, v_end((*source).vertices_end());
    double xmin, xmax, ymin, ymax, zmin, zmax;
    MyMesh::Point buffp;
    // Initialisation des variables
    xmax = -numeric_limits<double>::max();
    ymax = -numeric_limits<double>::max();
    zmax = -numeric_limits<double>::max();
    xmin = numeric_limits<double>::max();
    ymin = numeric_limits<double>::max();
    zmin = numeric_limits<double>::max();

    // Code principal
    for (v_it = (*source).vertices_begin() ; v_it != v_end ; ++v_it) {
        buffp = (*source).point(v_it);

        if (buffp[0] < xmin) {
            xmin = buffp[0];
        }

        if (buffp[0] > xmax) {
            xmax = buffp[0];
        }

        if (buffp[1] < ymin) {
            ymin = buffp[1];
        }

        if (buffp[1] > ymax) {
            ymax = buffp[1];
        }

        if (buffp[2] < zmin) {
            zmin = buffp[2];
        }

        if (buffp[2] > zmax) {
            zmax = buffp[2];
        }
    }

    (*p1)[0] = xmin;
    (*p1)[1] = ymin;
    (*p1)[2] = zmin;
    (*p2)[0] = xmax;
    (*p2)[1] = ymax;
    (*p2)[2] = zmax;
    cout << "xmin = " << xmin << endl;
    cout << "xmax = " << xmax << endl;
    cout << "ymin = " << ymin << endl;
    cout << "ymax = " << ymax << endl;
    cout << "zmin = " << zmin << endl;
    cout << "zmax = " << zmax << endl;
    cout << "Bounding box computed" << endl;
}


MyPolyMesh CreateBox(MyMesh::Point p1, MyMesh::Point p2) {
    MyPolyMesh mesh;
    // generate vertices
    MyPolyMesh::VertexHandle vhandle[8];
    vhandle[0] = mesh.add_vertex(MyPolyMesh::Point(p1[0], p1[1], p2[2]));
    vhandle[1] = mesh.add_vertex(MyPolyMesh::Point(p2[0], p1[1], p2[2]));
    vhandle[2] = mesh.add_vertex(MyPolyMesh::Point(p2[0], p2[1], p2[2]));
    vhandle[3] = mesh.add_vertex(MyPolyMesh::Point(p1[0], p2[1], p2[2]));
    vhandle[4] = mesh.add_vertex(MyPolyMesh::Point(p1[0], p1[1], p1[2]));
    vhandle[5] = mesh.add_vertex(MyPolyMesh::Point(p2[0], p1[1], p1[2]));
    vhandle[6] = mesh.add_vertex(MyPolyMesh::Point(p2[0], p2[1], p1[2]));
    vhandle[7] = mesh.add_vertex(MyPolyMesh::Point(p1[0], p2[1], p1[2]));
    // generate (quadrilateral) faces
    std::vector<MyPolyMesh::VertexHandle>  face_vhandles;
    face_vhandles.clear();
    face_vhandles.push_back(vhandle[0]);
    face_vhandles.push_back(vhandle[1]);
    face_vhandles.push_back(vhandle[2]);
    face_vhandles.push_back(vhandle[3]);
    mesh.add_face(face_vhandles);
    face_vhandles.clear();
    face_vhandles.push_back(vhandle[7]);
    face_vhandles.push_back(vhandle[6]);
    face_vhandles.push_back(vhandle[5]);
    face_vhandles.push_back(vhandle[4]);
    mesh.add_face(face_vhandles);
    face_vhandles.clear();
    face_vhandles.push_back(vhandle[1]);
    face_vhandles.push_back(vhandle[0]);
    face_vhandles.push_back(vhandle[4]);
    face_vhandles.push_back(vhandle[5]);
    mesh.add_face(face_vhandles);
    face_vhandles.clear();
    face_vhandles.push_back(vhandle[2]);
    face_vhandles.push_back(vhandle[1]);
    face_vhandles.push_back(vhandle[5]);
    face_vhandles.push_back(vhandle[6]);
    mesh.add_face(face_vhandles);
    face_vhandles.clear();
    face_vhandles.push_back(vhandle[3]);
    face_vhandles.push_back(vhandle[2]);
    face_vhandles.push_back(vhandle[6]);
    face_vhandles.push_back(vhandle[7]);
    mesh.add_face(face_vhandles);
    face_vhandles.clear();
    face_vhandles.push_back(vhandle[0]);
    face_vhandles.push_back(vhandle[3]);
    face_vhandles.push_back(vhandle[7]);
    face_vhandles.push_back(vhandle[4]);
    mesh.add_face(face_vhandles);
    return mesh;
}


double distance(MyMesh::Point *p, double x, double y, double z) {
    return sqrt(((*p)[0] - x) * ((*p)[0] - x) + ((*p)[1] - y) * ((*p)[1] - y) + ((*p)[2] - z) * ((*p)[2] - z));
}


double distance2(MyMesh::Point *p, MyMesh::Point *p2) {
    return sqrt(((*p)[0] - (*p2)[0]) * ((*p)[0] - (*p2)[0]) + ((*p)[1] - (*p2)[1]) * ((*p)[1] - (*p2)[1] ) + ((*p)[2] - (*p2)[2]) * ((*p)[2] - (*p2)[2]));
}


void DownSampling(MyMesh *source, MyMesh *res, int nb_vertices) {
    MyMesh::Point p1, p2, p;
    MyMesh::Point buffp;
    bool trouve = false;
    double incr, buff, dist;
    MyMesh::VertexIter v_it, v_end((*source).vertices_end());
    MyMesh::VertexIter v_it2, v_end2;
    int nb_vertices_res = 0;
    cout << "Downsampling the mesh" << endl;
    // initialisation factice de buffp pour eviter les warnings
    buffp = MyMesh::Point(0, 0, 0);

    if (nb_vertices > 2500) {
        // Partie uniforme
        BoundingBox(source, &p1, &p2);
        incr = (p2[0] - p1[0]) / 20.0;

        for (double x = p1[0] ; x < p2[0] ; x += incr) {
            for (double y = p1[1] ; y < p2[1] ; y += incr) {
                for (double z = p1[2] ; z < p2[2] ; z += incr) {
                    buff = numeric_limits<double>::max();

                    for (v_it = (*source).vertices_begin() ; v_it != v_end ; ++v_it) {
                        p = (*source).point(v_it);
                        dist = distance(&p, x, y, z);

                        if (dist < incr / 2 && dist < buff) {
                            buffp = p;
                            buff = dist;
                            trouve = true;
                        }
                    }

                    if (trouve) {
                        (*res).add_vertex(buffp);
                        trouve = false;
                    }
                }
            }
        }

        // Partie sur l'espace des normales
        // ou pas
    }

    v_end2 = (*res).vertices_end();

    for (v_it2 = (*res).vertices_begin() ; v_it2 != v_end2 ; ++v_it2) {
        nb_vertices_res++;
    }

    cout << "nb_vertices_res = " << nb_vertices_res << endl;
    cout << "Downsample done" << endl;
}


void G_distance(MyMesh *mesh, double*** res) {
    cout << "Computing geodesic distance between every points" << endl;
    int progress = 0;
    double mini;
    // Construction de la matrice d'adjacence
    MyMesh::VHandle v1, v2;
    double dist;
    int n = (*mesh).n_vertices();
    (*res) = (double **) malloc(n * sizeof(double*));

    for (int i = 0; i < n; i++) {
        (*res)[i] = (double*) malloc(n * sizeof(double));

        for (int j = 0; j <= i; j++) {
            v1 = MyMesh::VertexHandle(i);
            v2 = MyMesh::VertexHandle(j);

            if ((*mesh).find_halfedge (v1, v2) != MyMesh::InvalidVertexHandle || (*mesh).find_halfedge (v2, v1) != MyMesh::InvalidVertexHandle) {
                dist = distance2(&(*mesh).point(v1), &(*mesh).point(v2));
                (*res)[i][j] = dist;
                (*res)[j][i] = dist;
            } else {
                (*res)[i][j] = numeric_limits<double>::max();
                (*res)[j][i] = numeric_limits<double>::max();
            }

            if (i == j) {
                (*res)[i][j] = 0;
            }
        }
    }

    // Algorithme de Floyd Warshall (en place directement dans la matrice d'adjacence)
    cout << "Progression : 5% par #" << endl;
    cout << "[" << flush;

    for (int k = 0; k < n; k++) {
        if (k > (progress + 1) * n / 20) {
            progress = k * 20 / n;
            cout << "#" << flush;
        }

        for (int i = 0; i < n; i++) {
            for (int j = 0; j <= i; j++) {
                mini = min((*res)[i][j], (*res)[i][k] + (*res)[k][j]);
                (*res)[i][j] = mini;
                (*res)[j][i] = mini;
            }
        }
    }

    cout << "]" << endl;
    cout << "Geodesic distances computed" << endl;
//     for(uint i = 0; i < (*mesh).n_vertices(); i++) {
//       cout << "res[0][" << i << "] = " << (*res)[0][i] << endl;
//     }
}


void Neighborhoods(MyMesh *source, MyMesh::VHandle ***neighbors, double **distances) {
    int temp = 0;
    double dist = numeric_limits<double>::max();
    int n = (*source).n_vertices();
    cout << "Computing neighborhoods" << endl;
    (*neighbors) = (MyMesh::VHandle**)malloc(n * sizeof (MyMesh::VHandle*));

    for (int i = 0; i < n; i++) {
        (*neighbors)[i] = (MyMesh::VHandle*)malloc(nb_neighbors * sizeof (MyMesh::VHandle));

        for (int k = 0; k < nb_neighbors; k++) {
            for (int j = 0; j < n; j++) {
                if (distances[i][j] < dist && i != j) {
                    temp = j;
                    dist = distances[i][j];
                }
            }

            (*neighbors)[i][k] = MyMesh::VertexHandle(temp);
            distances[i][temp] = numeric_limits<double>::max(); // pour ne pas reprendre le meme voisin deux fois
            dist = numeric_limits<double>::max();
        }
    }

    cout << "Neighborhoods computed" << endl;
}


void Compute_Feature_V(MyMesh *mesh, MyMesh::VHandle vh, double *feature_v, double **distances, double d) {
    // Utilise la methode decrite dans : "Curvatures estimation on triangular mesh"
    list<MyMesh::VHandle> l1;
    list<MyMesh::VHandle> l2;
    list<MyMesh::VHandle> l3;
    list<MyMesh::VHandle> l4;
    double f1, f2;
    MyMesh::VertexIter v_it;
    MyMesh::VertexIter v_end = (*mesh).vertices_end();
//     cout << "d = " << d << endl;

    // remplissage des différentes listes, une par niveau
    for (v_it = (*mesh).vertices_begin() ; v_it != v_end ; ++v_it ) {
        double dist = distances[vh.idx()][MyMesh::VertexHandle(v_it).idx()];

//       cout << "dist = " << dist << endl;
        if (dist != 0 && dist < d / 25.0) {
            l4.push_back(MyMesh::VertexHandle(v_it));

            if (dist < d / 10.0) {
                l3.push_back(MyMesh::VertexHandle(v_it));

                if (dist < d / 15.0) {
                    l2.push_back(MyMesh::VertexHandle(v_it));

                    if (dist < d / 20.0) {
                        l1.push_back(MyMesh::VertexHandle(v_it));
                    }
                }
            }
        }
    }

//     cout << "l4.size() = " << l4.size() << endl;

    if (l1.size() == 0) {
        feature_v[0] = 0;
        feature_v[1] = 0;
    } else {
        compute_curv(mesh, vh, &l1, &f1, &f2);
        feature_v[0] = f1;
        feature_v[1] = f2;
    }

    if (l2.size() == 0) {
        feature_v[2] = 0;
        feature_v[3] = 0;
    } else {
        compute_curv(mesh, vh, &l2, &f1, &f2);
        feature_v[2] = f1;
        feature_v[3] = f2;
    }

    if (l3.size() == 0) {
        feature_v[4] = 0;
        feature_v[5] = 0;
    } else {
        compute_curv(mesh, vh, &l3, &f1, &f2);
        feature_v[4] = f1;
        feature_v[5] = f2;
    }

    if (l4.size() == 0) {
        feature_v[6] = 0;
        feature_v[7] = 0;
    } else {
        compute_curv(mesh, vh, &l4, &f1, &f2);
        feature_v[6] = f1;
        feature_v[7] = f2;
    }
}


void compute_curv(MyMesh *mesh, MyMesh::VHandle vh, list<MyMesh::VHandle> *liste, double *k1, double *k2) {
    gsl_vector *N = gsl_vector_calloc(3); // Normal estimated at the vertex
    MyMesh::VertexIter v_it;
    MyMesh::VertexIter v_end = (*mesh).vertices_end();
    MyMesh::Point pi;
    gsl_vector *p = gsl_vector_alloc(3);
    uint max_ind = 0;
    double max_kn = -numeric_limits<double>::max();
    gsl_vector *e1 = gsl_vector_calloc(3);
    gsl_vector *e2 = gsl_vector_calloc(3);
    list<MyMesh::VertexHandle>::iterator it;
    calculer_normales(mesh, vh, &N);
// cout << "N = (" << gsl_vector_get(N, 0) << ", " << gsl_vector_get(N, 1) << ", " << gsl_vector_get(N, 2) << ")" << endl;
    gsl_vector **Ni = (gsl_vector**)malloc((*liste).size() * sizeof(gsl_vector*));
    // calcul des normales pour chaque vertex autour du point p
    int i = 0;

    for (it = (*liste).begin(); it != (*liste).end(); ++it) {
        Ni[i] = gsl_vector_calloc(3);
        calculer_normales(mesh, *it, &(Ni[i]));
        i++;
    }

    // transformation du point p en un vecteur
    gsl_vector_set(p, 0, (*mesh).point(vh)[0]);
    gsl_vector_set(p, 1, (*mesh).point(vh)[1]);
    gsl_vector_set(p, 2, (*mesh).point(vh)[2]);
    // calcul des vecteurs (p-pi)
    gsl_vector **ppi = (gsl_vector**)malloc((*liste).size() * sizeof(gsl_vector*));
    i = 0;

    for (it = (*liste).begin(); it != (*liste).end(); ++it) {
        pi = (*mesh).point(*it);
        ppi[i] = gsl_vector_alloc(3);
        gsl_vector_set(ppi[i], 0, pi[0]);
        gsl_vector_set(ppi[i], 1, pi[1]);
        gsl_vector_set(ppi[i], 2, pi[2]);
        gsl_vector_sub(ppi[i], p);
        i++;
    }

    // calcul des projections des vecteurs (pi-p) sur le plan tangent
    gsl_vector **ti = (gsl_vector**)malloc((*liste).size() * sizeof(gsl_vector*));

    for (uint j = 0; j < (*liste).size(); j++) {
        ti[j] = gsl_vector_calloc(3);
        gsl_vector_add(ti[j], ppi[j]);
        gsl_vector *buff = gsl_vector_calloc(3);
        gsl_vector_add(buff, N);
        gsl_vector_scale(buff, gsl_vector_dot(ppi[j], N));
        gsl_vector_sub(ti[j], buff);
        gsl_vector_free(buff);
        gsl_vector_scale(ti[j], 1.0 / gsl_vector_norm(ti[j]));
//       cout << "ti(" << j << ") = (" << gsl_vector_get(ti[j], 0) << ", " << gsl_vector_get(ti[j], 1) << ", " << gsl_vector_get(ti[j], 2) << ")" << endl;
    }

    // calcul des courbures normales estimees
    double* kni = new double[(*liste).size()];

    for (uint j = 0; j < (*liste).size(); j++) {
        gsl_vector *buff = gsl_vector_calloc(3);
        gsl_vector_add(buff, Ni[j]);
        gsl_vector_sub(buff, N);
        kni[j] = -gsl_vector_dot(ppi[j], buff) / gsl_vector_dot(ppi[j], ppi[j]);

        if (kni[j] > max_kn) {
            max_kn = kni[j];
            max_ind = j;
        }

        gsl_vector_free(buff);
    }

//      cout << "nombre de sommets = " << (*liste).size() << endl;
//      cout << "max_ind = " << max_ind << endl;
//      cout << "ti[max_ind] = (" << gsl_vector_get(ti[max_ind],0) << ", " << gsl_vector_get(ti[max_ind],1) << ", " << gsl_vector_get(ti[max_ind],2) << ")" << endl;
    gsl_vector_add(e1, ti[max_ind]); // definition de e1
    gsl_vector_vect(e1, N, &e2); // definition de e2
    gsl_vector_scale(e2, 1.0 / gsl_vector_norm(e2));
//  cout << "e1 = (" << gsl_vector_get(e1, 0) << ", " << gsl_vector_get(e1, 1) << ", " << gsl_vector_get(e1, 2) << ")" << endl;
//  cout << "e2 = (" << gsl_vector_get(e2, 0) << ", " << gsl_vector_get(e2, 1) << ", " << gsl_vector_get(e2, 2) << ")" << endl;
    // calcul des angles delta[i] entre les vecteurs t[i] et le vecteur e1
    double* angle = new double[(*liste).size()];

    for (uint j = 0; j < (*liste).size(); j++) {
        if (j == max_ind) {
            angle[j] = 0;
        } else {
            angle[j] = acos((float)gsl_vector_dot(ti[j], e1));
// 	cout << "angle(" << j << ") = " << angle[j] << endl;
        }
    }

    // calcul des constantes de l'equation
    double a11, a12, a22, a13, a23;
    a11 = 0.0;
    a12 = 0.0;
    a22 = 0.0;
    a13 = 0.0;
    a23 = 0.0;

    for (uint j = 0; j < (*liste).size(); j++) {
        a11 += pow(cos(angle[j]), 2) * pow(sin(angle[j]), 2);
        a12 += cos(angle[j]) * pow(sin(angle[j]), 3);
        a22 += pow(sin(angle[j]), 4);
        a13 += kni[j] * (1 - pow(cos(angle[j]), 2)) * cos(angle[j]) * sin(angle[j]);
        a23 += kni[j] * (1 - pow(cos(angle[j]), 2)) * pow(sin(angle[j]), 2);
    }

    // calcul des constantes
    double a = max_kn;
    double b = (a13 * a22 - a23 * a12) / (a11 * a22 - pow(a12, 2));
    double c = (a11 * a23 - a12 * a13) / (a11 * a22 - pow(a12, 2));
    double Kg = a * c - pow(b, 2) / 4.0;
    double H = (a + c) / 2.0;
    *k1 = H + sqrt(pow(H, 2) - Kg);
    *k2 = H - sqrt(pow(H, 2) - Kg);
//  cout << "a = " << a << endl;
//  cout << "b = " << b << endl;
//  cout << "a11 = " << a11 << endl;
//  cout << "a22 = " << a22 << endl;
//  cout << "a12 = " << a12 << endl;
//  cout << "c = " << c << endl;
//  cout << "Kg = " << Kg << endl;
//  cout << "H = " << H << endl;
//  cout << "k1 = " << *k1 << endl;
//  cout << "k2 = " << *k2 << endl;

    // liberation de la memoire
    for (uint j = 0; j < (*liste).size(); j++) {
        gsl_vector_free(Ni[j]);
        gsl_vector_free(ti[j]);
        gsl_vector_free(ppi[j]);
    }

    gsl_vector_free(N);
    free(Ni);
    free(ti);
    free(ppi);
    gsl_vector_free(p);
    gsl_vector_free(e1);
    gsl_vector_free(e2);
    delete[] kni;
}


void gslize(MyMesh::Normal *norm, gsl_vector *N) {
    gsl_vector_set(N, 0, (double)(*norm).data()[0]);
    gsl_vector_set(N, 1, (double)(*norm).data()[1]);
    gsl_vector_set(N, 2, (double)(*norm).data()[2]);
}


double gsl_vector_dot(gsl_vector* v1, gsl_vector* v2) {
    return gsl_vector_get(v1, 0) * gsl_vector_get(v2, 0) + gsl_vector_get(v1, 1) * gsl_vector_get(v2, 1) +  gsl_vector_get(v1, 2) * gsl_vector_get(v2, 2);
}


double gsl_vector_norm(gsl_vector* v) {
    return sqrt(gsl_vector_dot(v, v));
}


void gsl_vector_vect(gsl_vector* v1, gsl_vector* v2, gsl_vector** res) {
    if (*res == NULL) {
        *res = gsl_vector_alloc(3);
    }

    gsl_vector_set(*res, 0, gsl_vector_get(v1, 1) * gsl_vector_get(v2, 2) - gsl_vector_get(v1, 2) * gsl_vector_get(v2, 1));
    gsl_vector_set(*res, 1, gsl_vector_get(v1, 2) * gsl_vector_get(v2, 0) - gsl_vector_get(v1, 0) * gsl_vector_get(v2, 2));
    gsl_vector_set(*res, 2, gsl_vector_get(v1, 0) * gsl_vector_get(v2, 1) - gsl_vector_get(v1, 1) * gsl_vector_get(v2, 0));
}


void calculer_normales(MyMesh *mesh, MyMesh::VHandle vh, gsl_vector **N) {
    MyMesh::VertexFaceIter vf_it; // circulateur sur les faces autours du point
    MyMesh::FaceVertexIter fv_it; // circulateur sur les points constituant une face
    gsl_vector *gi = gsl_vector_calloc(3);
    gsl_vector *p = gsl_vector_calloc(3);
    gsl_vector *pj = gsl_vector_calloc(3);
    gsl_vector *p0 = gsl_vector_calloc(3);
    gsl_vector *p1 = gsl_vector_calloc(3);
    gsl_vector *p2 = gsl_vector_calloc(3);
    double wi;
    gsl_vector *Nf = gsl_vector_calloc(3);
    MyMesh::Point* ppj = new MyMesh::Point[3];
    int i;
    // création du vecteur representant le point p
    gsl_vector_set(p, 0, (*mesh).point(vh)[0]);
    gsl_vector_set(p, 1, (*mesh).point(vh)[1]);
    gsl_vector_set(p, 2, (*mesh).point(vh)[2]);

    // pour chaque face autour du point
    for (vf_it = (*mesh).vf_iter(vh); vf_it; ++vf_it) {
        i = 0;

        for (fv_it = (*mesh).fv_iter(vf_it); fv_it; ++fv_it) {
            ppj[i] = (*mesh).point(fv_it); // On stocke les 3 points de la face
            i++;
        }

        // calcul de gi
        for (int j = 0; j < 3; j++) {
            gsl_vector_set(pj, 0, ppj[j][0]);
            gsl_vector_set(pj, 1, ppj[j][1]);
            gsl_vector_set(pj, 2, ppj[j][2]);
            gsl_vector_add(gi, pj);
            gsl_vector_scale(gi, 1.0 / 3.0);
        }

        // calcul de wi
        gsl_vector_sub(gi, p);
        wi = 1.0 / gsl_vector_norm(gi);

        // calcul de Nf
        for (int j = 0; j < 3; j++) {
            gsl_vector_set(p0, 0, ppj[0][0]);
            gsl_vector_set(p0, 1, ppj[0][1]);
            gsl_vector_set(p0, 2, ppj[0][2]);
            gsl_vector_set(p1, 0, ppj[1][0]);
            gsl_vector_set(p1, 1, ppj[1][1]);
            gsl_vector_set(p1, 2, ppj[1][2]);
            gsl_vector_set(p2, 0, ppj[2][0]);
            gsl_vector_set(p2, 1, ppj[2][1]);
            gsl_vector_set(p2, 2, ppj[2][2]);
        }

        gsl_vector_sub(p1, p0); // vecteur du point 0 au point 1
        gsl_vector_sub(p2, p0);// vecteur du point 0 au point 2
        gsl_vector_vect(p1, p2, &Nf);
        gsl_vector_scale(Nf, wi / gsl_vector_norm(Nf));
        // ajout de wi*Nf a N
        gsl_vector_add(*N, Nf);
        // remise a zero de gi pour la prochaine boucle
        gsl_vector_set(gi, 0, 0);
        gsl_vector_set(gi, 1, 0);
        gsl_vector_set(gi, 2, 0);
    }

    // normalisation de N
    gsl_vector_scale(*N, 1.0 / gsl_vector_norm(*N));
    gsl_vector_free(gi);
    gsl_vector_free(p);
    gsl_vector_free(pj);
    gsl_vector_free(p0);
    gsl_vector_free(p1);
    gsl_vector_free(p2);
}


}

