#ifndef PREPROCESSING_H
#define PREPROCESSING_H

#include "icp2.h"

// -------------------- OpenMesh
#include <OpenMesh/Core/IO/MeshIO.hh>
#include <OpenMesh/Core/Mesh/Types/TriMesh_ArrayKernelT.hh>
#include <OpenMesh/Core/Mesh/Types/PolyMesh_ArrayKernelT.hh>

typedef OpenMesh::PolyMesh_ArrayKernelT<>  MyPolyMesh;
typedef OpenMesh::TriMesh_ArrayKernelT<> MyMesh;

namespace Preprocessing {

void BoundingBox(MyMesh *source, MyMesh::Point *p1, MyMesh::Point *p2);
// Compute the 2 points describing the bounding box

MyPolyMesh CreateBox(MyMesh::Point p1, MyMesh::Point p2);
// Create a box from 2 given points

double distance(MyMesh::Point *p, double x, double y, double z);
// Compute the distance between a point p and a point (x, y ,z)

double distance2(MyMesh::Point *p, MyMesh::Point *p2);

void DownSampling(MyMesh *source, MyMesh *res, int nb_vertices);
// Downsample a mesh

void Neighborhoods(MyMesh *source, MyMesh::VHandle ***neighbors, double **distances);
// Compute a array of points which are the "nb_neighbors" nearest neighbors of each point of the mesh source

void G_distance(MyMesh *mesh, double*** res);
// Compute the geodesic distance of two points on a mesh

void gslize(MyMesh::Normal *norm, gsl_vector *N);

double gsl_vector_dot(gsl_vector* v1, gsl_vector* v2);

double gsl_vector_norm(gsl_vector* v);

void gsl_vector_vect(gsl_vector* v1, gsl_vector* v2, gsl_vector** res);

void calculer_normales(MyMesh *mesh, MyMesh::VHandle vh, gsl_vector **N);

void Compute_Feature_V(MyMesh *mesh, MyMesh::VHandle vh, double *feature_v, double **distances, double d);
// Compute the feature vector of a mesh at point p1

void compute_curv(MyMesh *mesh, MyMesh::VHandle vh, std::list<MyMesh::VHandle> *liste, double *k1, double *k2);

}
#endif
