#ifndef CLUSTER_H
#define CLUSTER_H

#include "Correspondence.h"

//using namespace std;
//using namespace icp2;

typedef OpenMesh::TriMesh_ArrayKernelT<>  MyMesh;

struct _Cluster;
struct _Sample;

typedef _Sample Sample;
typedef _Cluster Cluster;

struct _Cluster {
    int name;
    //proprietes de chaque point du cluster
    std::list<Sample> V_set;
    //pts du cluster
    std::list<MyMesh::VertexHandle> pts_source;
    //points de la cible associes
    std::list<MyMesh::VertexHandle> pts_cible;
    //liste des clusters voisins
    std::list<Cluster*> clust_voisins;
    // Rotation et Translation optimum
    double ** transfo;
    double erreur;
    int traiter;
    // calcul des couleurs
    int* ind_color;  // ind_color[0] : couleur du cluster, ind_color[1..4] : couleur des voisins (0 ou 1)
};

struct _Sample {
    // Correspondence
    MyMesh::VertexHandle *source_subset;
    MyMesh::VertexHandle *source_prim_subset;
    MyMesh::VertexHandle *target;
    // list of extended clusters that contain the sample
    std::list<Cluster*> Clusters_tild;
    double w;
};

namespace Clusters {

/* fonction qui initialise les clusters : chaque cluster est constitué d'un point pour le moment
 *
 * La fonction prend en entrée:
 *       - un pointeur sur une liste (vide au départ, la liste est rempli par la fonction)
 *       - Les deux maillages source et cible, ainsi que les tableaux de leurs points (VertexHandle)
 *       - La taille des tableaux associés (les deux tableaux ont la même taille "nb_vertices"
 *       - Les plus proches voisins des points et leurs correspondances
 *       - Le nombre de voisins que l'on qualifie de "plus proches"
 *
 */
void init_clusters(std::list<Cluster*>** ls_clust, MyMesh* source, MyMesh::VertexHandle** pts_source, MyMesh* cible, MyMesh::VertexHandle** pts_cible, MyMesh::VertexHandle*** correspondence, MyMesh::VertexHandle*** neighbors, int nb_vertices, int nb_neighbors);

/* fonction qui effectue le traitement selon l'algorithme defini
 *
 * à la fin, ls_clust renvoie les clusters fusionnés en fonction du seuil E_max défini
 *
 */
void process_clustering(std::list<Cluster*>* ls_clust, MyMesh* source, MyMesh::VertexHandle** pts_src, MyMesh* cible, MyMesh::VertexHandle** pts_target, MyMesh::VertexHandle*** neighbors, MyMesh::VertexHandle*** correspondence, int nb_neighbors, double E_max);

/*fonction qui trie les clusters selon le paramètre "erreur" de la struture cluster
 */
void tri_err(std::list<Cluster*>* ls_clust);

/*fonction qui fusionne une liste de cluster et le stocke dans le cluster "resultat"
 */
void fusion(std::list<Cluster*>* ls_clust, Cluster* resultat, MyMesh::VertexHandle*** neighbors, MyMesh::VertexHandle*** correspondence, MyMesh* source, MyMesh* cible, int nb_neighbors);

/* fonction qui renvoie true si un cluster c appartient à une liste de cluster
 */
bool appartient(Cluster* c, std::list<Cluster*>* ls_clust);

/* fonction qui a un cluster renvoie la liste des points de la source qui sont dans le cluster étendu
 */
void cluster_tilde(Cluster* c, MyMesh::VertexHandle*** neighbours, std::list<MyMesh::VertexHandle>* resultat, int nb_neighbors);
/*
 * fonction qui calcule les couleurs des clusters
 */
void color_clusters(std::list<Cluster*>* ls_clust, const int nb_color, int** couleurs, const int vsize);
}

#endif
