#include "energy.h"

using namespace std;
using namespace icp2;
using namespace Clusters;

namespace energy {

void Init_E(MyMesh *source_subset, MyMesh *source_prim_subset, MyMesh *target,
            MyMesh::VertexHandle*** neighbours,
            list<Sample*>** s, list<Cluster*>* clusters,
            MyMesh::VertexHandle** corresp, double* poids) {
    //printf("Enter Init_E\n");
    Sample* se;
    int i;
    *s = new list<Sample*>;

    // init s
    for (i = 0; i < source_subset->n_vertices(); ++i) {
        se = new Sample;
        se->source_subset = corresp[0] + i;
        se->source_prim_subset = corresp[0] + i;
        se->target = corresp[1] + i;
        se->w = poids[i];
        (*s)->push_back(se);
    }

    // init clusters_tild
    list<Sample*>::iterator it_s;
    list<Cluster*>::iterator it_k;
    list<MyMesh::VertexHandle> v;
    list<MyMesh::VertexHandle>::iterator it_v;
    int nb_neighbors = 15;

    for (it_k = clusters->begin(); it_k != clusters->end(); ++it_k) {
        cluster_tilde(*it_k, neighbours, &v, nb_neighbors);

        for (it_v = v.begin(); it_v != v.end(); ++it_v) {
            for (it_s = (*s)->begin(); *((*it_s)->source_subset) != *it_v; ++it_s);

            (*it_s)->Clusters_tild.push_back(*it_k);
        }
    }

    //printf("Leave Init_E\n");
}

double Minimise_E(MyMesh *source_subset, MyMesh *source_prim_subset, MyMesh *target,
                  list<Sample*>* s, list<Cluster*>* clusters) {
    cout << "Energy minimization" << endl;
    const bool test = false;
    const int max_iter = 10;
    const double alpha = 0.6, beta = 0.4;
    const int l_c = 2, l_r = 1;
    MyMesh *source = source_subset;
    double E, E_c, E_r;
    double tmp;
    int iter;
    Cluster* cl;
    MyMesh::VertexHandle *source_h, *s_prim_h, *target_h;
    int vsize;
    int i, k;
    gsl_matrix *D, *N, *R, *R_k, *M, *A;
    gsl_vector *s_i, *s2_i, *t_i, *n_i, *T, *T_k, *b;
    list<Sample*>::iterator it_s;
    list<Cluster*>::iterator it_k;
    list<Sample>::iterator it_s2;
    Sample samp;
    int idx;
    MyMesh::Point p;
    MyMesh::Point* pp;
    D = gsl_matrix_alloc(3, 3);
    N = gsl_matrix_alloc(3, 3);
    R = gsl_matrix_alloc(3, 3);
    R_k = gsl_matrix_alloc(3, 3);
    M = gsl_matrix_alloc(3, 3);
    A = gsl_matrix_alloc(3, 3);
    s_i = gsl_vector_alloc(3);
    s2_i = gsl_vector_alloc(3);
    t_i = gsl_vector_alloc(3);
    n_i = gsl_vector_alloc(3);
    T = gsl_vector_alloc(3);
    T_k = gsl_vector_alloc(3);
    b = gsl_vector_alloc(3);
    cout << "[" << flush;

    for (iter = 0; iter < max_iter; iter++) {
        E_r = 0;
        E_c = 0;
        /* Minimisation de E

        E = l_c.E_c + l_r.E_r

        E_c = SUM[s_i in S_subset] w_i.(alpha.|s'_i - t_i|² + beta.((s'_i - t_i).n_i)²)

        E_r = SUM[k] SUM[s_i in C_tild[k]] |R_k.s_i + T_k - s'_i|²

        */
        //////////////////////////////////////////////////////////////////
        // minization wrt s'_i
//       for (it_s = s->begin(); it_s != s->end(); ++it_s){
// 	samp = **it_s;
// 	p = source_subset->point(*(samp.source_subset));
// 	cout<<"source sample ";
// 	cout<<samp.source_subset->idx();
// 	cout<<"\t"<<p[0]<<" "<<p[1]<<" "<<p[2]<<endl;
// 	p = source_prim_subset->point(*(samp.source_prim_subset));
// 	cout<<"s prim sample ";
// 	cout<<samp.source_prim_subset->idx();
// 	cout<<"\t"<<p[0]<<" "<<p[1]<<" "<<p[2]<<endl;
// 	p = target->point(*(samp.target));
// 	cout<<"target sample ";
// 	cout<<samp.target->idx();
// 	cout<<"\t"<<p[0]<<" "<<p[1]<<" "<<p[2]<<endl;
// 	cout << endl;
//       }
        const int version = 3;

        for (it_s = s->begin(); it_s != s->end(); ++it_s) {
            if (version == 0) {
                // s'_i = (SUM[C_tild_i] (R_k.s_i + T_k))/#[C_tild_i]
                // s2_i = SUM[C_tild_i] (R_k.s_i + T_k)
                samp = **it_s;
                gsl_vector_set_zero(s2_i);

                for (it_k = samp.Clusters_tild.begin(); it_k != samp.Clusters_tild.end(); ++it_k) {
                    Rot((*it_k)->transfo, R_k);
                    Trans((*it_k)->transfo, T_k);
                    Mem_Copy_Point_V((*source).point(*(samp.source_subset)), s_i, 3);
                    Mult_Matrice_Vector(R_k, s_i, 3);
                    gsl_vector_add(s_i, T_k);
                    gsl_vector_add(s2_i, s_i);
                }

                // s2_i /= #[C_tild_i]
                tmp = samp.Clusters_tild.size();
                tmp = 1 / tmp;
                gsl_vector_scale(s2_i, tmp);

                // s'_i = s2_i
                for (k = 0; k < 3; ++k) {
                    ((*source_prim_subset).point(*((*it_s)->source_prim_subset)))[k] = gsl_vector_get (s2_i, k);
                }
            } else if (version == 1) {
                // t_i = t_i
                Mem_Copy_Point_V((*target).point(*((*it_s)->target)), t_i, 3);
                // s2_i = t_i
                Mem_Copy_Point_V((*target).point(*((*it_s)->target)), s2_i, 3);

                for (k = 0; k < 3; ++k) {
                    ((*source_prim_subset).point(*((*it_s)->source_prim_subset)))[k] = gsl_vector_get (s2_i, k);
                }
            } else if (version == 2) {
                samp = **it_s;
                // A = (l_c.alpha.w_i + l_r.#).I
                tmp = samp.Clusters_tild.size();
                tmp = l_r * alpha * samp.w + l_r * tmp;
                gsl_matrix_set_identity(A);
                gsl_matrix_scale(A, tmp);
                //aff_mat(A,"A");
                // b = SUM[C_tild_i] (R_k.s_i + T_k)
                gsl_vector_set_zero(b);

                for (it_k = samp.Clusters_tild.begin(); it_k != samp.Clusters_tild.end(); ++it_k) {
                    Rot((*it_k)->transfo, R_k);
                    Trans((*it_k)->transfo, T_k);
                    Mem_Copy_Point_V((*source).point(*(samp.source_subset)), s_i, 3);
                    Mult_Matrice_Vector(R_k, s_i, 3);
                    gsl_vector_add(s_i, T_k);
                    gsl_vector_add(b, s_i);
                }

                // b = l_c.alpha.w_i.t_i + l_r.b
                gsl_vector_scale(b, l_r);
                Mem_Copy_Point_V((*target).point(*(samp.target)), t_i, 3);
                gsl_vector_scale(t_i, l_c*alpha*samp.w);
                gsl_vector_add(b, t_i);
                //aff_vec(b,"b");
                // A.s2_i = b
                gsl_linalg_HH_solve(A, b, s2_i);

                for (k = 0; k < 3; ++k) {
                    ((*source_prim_subset).point(*(samp.source_prim_subset)))[k] = gsl_vector_get (s2_i, k);
                }
            } else if (version == 3) {
                // normal vector computing
                Vect_Normal((*it_s)->target, target, n_i);
                // D = alpha.I_3
                D = gsl_matrix_alloc(3, 3);
                gsl_matrix_set_identity(D);
                gsl_matrix_scale(D, alpha);
                // N = n_i.n_i^T
                Mult_Matrice(n_i, n_i, &N);
                // R = SUM(R_k), T = SUM(T_k)
                gsl_matrix_set_zero(R);
                gsl_vector_set_zero(T);

                for (it_k = (*it_s)->Clusters_tild.begin(); it_k != (*it_s)->Clusters_tild.end(); ++it_k) {
                    Rot((*it_k)->transfo, R_k);
                    Trans((*it_k)->transfo, T_k);
                    gsl_matrix_add(R, R_k);
                    gsl_vector_add(T, T_k);
                }

                // M = l_c.w_i.(D + beta.N)
                gsl_matrix_memcpy(M, N);
                gsl_matrix_scale(M, beta);
                gsl_matrix_add(M, D);
                gsl_matrix_scale(M, l_c*((*it_s)->w));
                // A = M + l_r.#{k|s_i in C_tild_k}.I_3
                gsl_matrix_set_identity(A);
                gsl_matrix_scale(A, l_r*((*it_s)->Clusters_tild.size()));
                gsl_matrix_add(A, M);
                // b = M.t_i + l_r.(R.s_i + T)
                Mem_Copy_Point_V((*source).point(*((*it_s)->source_subset)), b, 3);
                Mult_Matrice_Vector(R, b, 3);
                gsl_vector_add(b, T);
                gsl_vector_scale(b, l_r);
                Mem_Copy_Point_V((*target).point(*((*it_s)->target)), t_i, 3);
                Mult_Matrice_Vector(M, t_i, 3);
                gsl_vector_add(b, t_i);
                // s2_i = A^-1.b
                gsl_linalg_HH_solve(A, b, s2_i);

                // s'_i = s2_i
                for (k = 0; k < 3; ++k) {
                    ((*source_prim_subset).point(*((*it_s)->source_prim_subset)))[k] = gsl_vector_get (s2_i, k);
                }
            }

            // E_c += w_i [alpha.norm(s'_i-t_i)^2]
            Mem_Copy_Point_V((*target).point(*((*it_s)->target)), t_i, 3);
            gsl_vector_sub(s2_i, t_i);
            E_c += ((*it_s)->w) * (alpha * Scalar(s2_i, s2_i));
            //	cout << "E_c : " << E_c << endl;
        }

        ///////////////////////////////////////////////////////////////
        // minimization wrt R_k, T_k

        for (it_k = clusters->begin(); it_k != clusters->end(); ++it_k) {
            cl = *it_k;
            vsize = cl->V_set.size();
            source_h = new MyMesh::VertexHandle[vsize];
            s_prim_h = new MyMesh::VertexHandle[vsize];
            target_h = new MyMesh::VertexHandle[vsize];
            i = 0;

            for (it_s2 = cl->V_set.begin(); it_s2 != cl->V_set.end(); ++it_s2) {
                source_h[i] = *((*it_s2).source_subset);
                s_prim_h[i] = *((*it_s2).source_prim_subset);
                target_h[i] = *((*it_s2).target);
                ++i;
            }

            Align(source_prim_subset, s_prim_h, target, target_h, &(cl->transfo), &tmp, vsize);

            if (iter == max_iter - 1 || test) {
                E_r += tmp;
            }

            //cout << "E_r : " << E_r << endl;
            free(source_h);
            free(s_prim_h);
            free(target_h);
// 	for (it_s2 = cl->V_set.begin(); it_s2 != cl->V_set.end(); ++it_s2){
// 	  // s'_i = R_k.s_i + T_k
// 	  Mem_Copy_Point_V((*source_subset).point(*((*it_s2).source_subset)), s_i, 3);
// 	  Rot(cl->transfo,R_k);
//  	  Trans(cl->transfo,T_k);
// 	  Mult_Matrice_Vector(R_k,s_i,3);
//  	  gsl_vector_add(s_i,T_k);
// 	  for (k = 0; k < 3; ++k){
// 	    ((*source_subset).point(*((*it_s2).source_subset)))[k] = gsl_vector_get(s_i,k);
// 	  }
// 	}
        }

//       if (iter == max_iter -1 || test){
// 	// energy update
// 	E = l_r * E_r + l_c * E_c;
// 	cout << "E : " << E << endl;
//       }

        if (iter > 0) {
            free(source);
        }

        source = new MyMesh;
        *source = MyMesh(*source_prim_subset);
        cout << "##" << flush;
    }

    cout << "]" << endl;
    gsl_matrix_free(D);
    gsl_matrix_free(N);
    gsl_matrix_free(R);
    gsl_matrix_free(R_k);
    gsl_matrix_free(M);
    gsl_matrix_free(A);
    gsl_vector_free(s_i);
    gsl_vector_free(s2_i);
    gsl_vector_free(t_i);
    gsl_vector_free(n_i);
    gsl_vector_free(T);
    gsl_vector_free(T_k);
    gsl_vector_free(b);
    return E;
};




//////////////////////////
// Procedures auxiliaires
//////////////////////////

void Vect_Normal(MyMesh::VertexHandle *vh, MyMesh *mesh, gsl_vector *res) {
    //printf("Enter Vect_Normal\n");
    Preprocessing::calculer_normales(mesh, *vh, &res);
    //printf("Leave Vect_Normal\n");
};

void Rot(double **transfo, gsl_matrix *rot) {
    int i, j;

    for (i = 0; i < 3; ++i) {
        for (j = 0; j < 3; ++j) {
            gsl_matrix_set (rot, i, j, transfo[i][j]);
        }
    }
};

void Trans(double **transfo, gsl_vector *transla) {
    int i;

    for (i = 0; i < 3; ++i) {
        gsl_vector_set (transla, i, transfo[i][3]);
    }
};

double Scalar(const gsl_vector *v1, const gsl_vector *v2) {
    double res = 0;
    int i;

    for (i = 0; i < 3; ++i) {
        res += gsl_vector_get(v1, i) * gsl_vector_get(v2, i);
    }

    return res;
}

void aff_mat(gsl_matrix *M, const char* name) {
    int i, j;
    printf("%s : \n", name);

    for (i = 0; i < 3; ++i) {
        printf("\t");

        for (j = 0; j < 3; ++j) {
            printf(" %f", gsl_matrix_get(M, i, j));
        }

        cout << endl;
    }
}

void aff_vec(gsl_vector *V, const char* name) {
    int i;
    printf("%s :\t", name);

    for (i = 0; i < 3; ++i) {
        printf(" %f", gsl_vector_get(V, i));
    }

    cout << endl;
}

}

