#ifndef ENERGY_H
#define ENERGY_H

#include "cluster.h"

namespace energy {

/** Initialisation de la liste des appartenances de chaque sample
 * i.e. les Clusters_tild (du type Sample, defini dans cluster.h)
 */
void Init_E(MyMesh *source_subset, MyMesh *source_prim_subset, MyMesh *target,
            MyMesh::VertexHandle*** neighbours,
            std::list<Sample*>** s, std::list<Cluster*>* clusters,
            MyMesh::VertexHandle** corresp, double* poids);

/** Fonction principale
 * Retourne l'energie finale
 * w est la liste des poids -> eq. (6) de [HAWG08] (Bertrand)
 */
double Minimise_E(MyMesh *source_subset, MyMesh *source_prim_subset, MyMesh *target,
                  std::list<Sample*>* s, std::list<Cluster*>* clusters);


/** Fonctions annexes **/

// Calcul de normal
// ne fait aucune allocation
void Vect_Normal(MyMesh::VertexHandle *V, MyMesh *mesh, gsl_vector *res);

// Converti une transformation 4x4 en matrice de rotation
// ne fait aucune allocation
void Rot(double **transfo, gsl_matrix *rot);

// Converti une transformation 4x4 en une vecteur translation
// ne fait aucune allocation
void Trans(double **transfo, gsl_vector *transla);

// Renvoie la somme du produit des coordonnees
double Scalar(const gsl_vector *v1, const gsl_vector *v2);

// Affiche
void aff_mat(gsl_matrix *M, const char* name);
void aff_vec(gsl_vector *V, const char* name);
}

#endif
