#include "icp2.h"

using namespace std;

namespace icp2 {


void Align (MyMesh *source, MyMesh::VertexHandle *source_handle, MyMesh *model, MyMesh::VertexHandle *model_handle, double*** transformation_out, double *erreur, int nb_vertices) {
    // a decommenter pour avoir les traces
    //#define Verbose
    gsl_vector * up = NULL;
    gsl_vector * ux = NULL;
    Center_Mass(source, source_handle, &up, nb_vertices);
    Center_Mass(model, model_handle, &ux, nb_vertices);
#ifdef Verbose
    printf("Center of mass of source: \n");
    Display_Vector(up, 3);
    printf("Center of mass of model: \n");
    Display_Vector(ux, 3);
#endif
    gsl_matrix * sigma_px = NULL;
    Cross_Covariance(source, source_handle, model, model_handle, up, ux, &sigma_px, nb_vertices);
#ifdef Verbose
    printf("Cross covariance matrix: \n");
    Display_Matrix(sigma_px, 3, 3);
#endif
    gsl_vector * delta = NULL;
    Delta_M(sigma_px, &delta);
#ifdef Verbose
    printf("Delta vector:\n");
    Display_Vector(delta, 3);
#endif
    gsl_matrix * Q = NULL;
    Q_Matrix(sigma_px, delta, &Q);
#ifdef Verbose
    printf("Q matrix:\n");
    Display_Matrix(Q, 4, 4);
#endif
    gsl_vector * qr = NULL;
    //Calcul du vecteur propre associe a la valeur propre max de la matrice Q
    Eigen_V(Q, &qr, 4);
#ifdef Verbose
    printf("Eigen Vector max (qr):\n");
    Display_Vector(qr, 4);
#endif
    gsl_matrix * Rqr = NULL;
    Q_Rotation(qr, &Rqr);
#ifdef Verbose
    printf("R(qr) matrix:\n");
    Display_Matrix(Rqr, 3, 3);
#endif
    gsl_vector * qt = gsl_vector_alloc (3);
    gsl_vector_memcpy (qt, ux);
    // Attention up est remplace apres cette operation
    Mult_Matrice_Vector(Rqr, up, 3);
    gsl_vector_sub (qt, up);
#ifdef Verbose
    printf("Optimal translation vector:\n");
    Display_Vector(qt, 3);
#endif
    //Resultat
    (*transformation_out) = (double**) malloc(4 * sizeof(double));

    for (int i = 0; i < 4; i++) {
        (*transformation_out)[i] = (double*) malloc(4 * sizeof(double));
    }

    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            (*transformation_out)[i][j] = (double) gsl_matrix_get (Rqr, i, j);
        }
    }

    for (int i = 0; i < 3; i++) {
        (*transformation_out)[i][3] = (double) gsl_vector_get(qt, i);
        (*transformation_out)[3][i] = 0;
    }

    (*transformation_out)[3][3] = 1;
    //Affichage
#ifdef Verbose
    printf("Final Result:\n");
    Display_Transf (*transformation_out);
#endif
    *erreur = mean_square_error(source, source_handle, model, model_handle, *transformation_out, nb_vertices);
    gsl_vector_free (qt);
    gsl_matrix_free (Rqr);
    gsl_vector_free (ux);
    gsl_vector_free (up);
    gsl_matrix_free (sigma_px);
    gsl_vector_free (delta);
    gsl_matrix_free (Q);
    gsl_vector_free (qr);
}



void Center_Mass(MyMesh *mesh, MyMesh::VertexHandle *x, gsl_vector ** u, int nb_vertices) {
    *u = gsl_vector_alloc (3);
    MyMesh::Point p;
    gsl_vector * v = gsl_vector_alloc (3);
    gsl_vector_set_zero (*u);
    gsl_vector_set_zero (v);

    for (int i = 0; i < nb_vertices; i++) {
        p = Access_Point(mesh, x[i]);
        Mem_Copy_Point_V(p, v, 3);
        gsl_vector_add (*u, v);
    }

    gsl_vector_scale (*u, (double)(1 / (double)nb_vertices));
    gsl_vector_free (v);
}

void Cross_Covariance(MyMesh *p, MyMesh::VertexHandle *p_handle, MyMesh *x, MyMesh::VertexHandle *x_handle, gsl_vector * up, gsl_vector * ux, gsl_matrix ** sigma_px, int nb_vertices) {
    *sigma_px = gsl_matrix_alloc (3, 3);
    gsl_vector * u = gsl_vector_alloc (3);
    gsl_vector * v = gsl_vector_alloc (3);
    gsl_vector_set_zero (u);
    gsl_vector_set_zero (u);
    gsl_matrix * m;
    gsl_matrix_set_zero (*sigma_px);
    MyMesh::Point pt;

    for (int i = 0; i < nb_vertices; i++) {
        pt = Access_Point(p, p_handle[i]);
        Mem_Copy_Point_V(pt, u, 3);
        gsl_vector_sub (u, up);
        pt = Access_Point(x, x_handle[i]);
        Mem_Copy_Point_V(pt, v, 3);
        gsl_vector_sub (v, ux);
        Mult_Matrice(u, v, &m);
        gsl_matrix_add (*sigma_px, m);
    }

    gsl_matrix_scale (*sigma_px, (double)(1 / (double)nb_vertices));
    gsl_matrix_free (m);
    gsl_vector_free (u);
    gsl_vector_free (v);
}

void Delta_M(gsl_matrix * m, gsl_vector ** delta) {
    *delta = gsl_vector_alloc (3);
    gsl_vector_set_zero (*delta);
    gsl_matrix * A = gsl_matrix_alloc (3, 3);
    gsl_matrix * A_transpose = gsl_matrix_alloc (3, 3);
    gsl_matrix_memcpy (A, m);
    gsl_matrix_transpose_memcpy (A_transpose, A);
    gsl_matrix_sub (A, A_transpose);
    double cour;
    //delta[0]=A23
    cour = gsl_matrix_get (A, 1, 2);
    gsl_vector_set(*delta, 0, cour);
    //delta[1]=A31
    cour = gsl_matrix_get (A, 2, 0);
    gsl_vector_set(*delta, 1, cour);
    //delta[2]=A12
    cour = gsl_matrix_get (A, 0, 1);
    gsl_vector_set(*delta, 2, cour);
    gsl_matrix_free (A);
    gsl_matrix_free (A_transpose);
}

void Q_Matrix(gsl_matrix * sigma_px, gsl_vector * delta, gsl_matrix ** q_matrix) {
    *q_matrix = gsl_matrix_alloc (4, 4);
    gsl_matrix_set_zero (*q_matrix);
    double trace = gsl_matrix_get (sigma_px, 0, 0) + gsl_matrix_get (sigma_px, 1, 1) + gsl_matrix_get (sigma_px, 2, 2);
    gsl_matrix * Id = gsl_matrix_alloc (3, 3);
    gsl_matrix_set_zero (Id);

    for (int i = 0; i < 3; i++) {
        gsl_matrix_set (Id, i, i, 1);
    }

    gsl_matrix_scale (Id, trace);
    gsl_matrix * A = gsl_matrix_alloc (3, 3);
    gsl_matrix * A_transpose = gsl_matrix_alloc (3, 3);
    gsl_matrix_memcpy (A, sigma_px);
    gsl_matrix_transpose_memcpy (A_transpose, A);
    gsl_matrix_add (A, A_transpose);
    gsl_matrix_sub (A, Id);
    //Construction of the matrix Q
    gsl_matrix_set (*q_matrix, 0, 0, trace);

    for (int i = 1; i < 4; i++) {
        gsl_matrix_set (*q_matrix, 0, i, gsl_vector_get(delta, i - 1));
        gsl_matrix_set (*q_matrix, i, 0, gsl_vector_get(delta, i - 1));
    }

    for (int i = 1; i < 4; i++) {
        for (int j = 1; j < 4; j++) {
            gsl_matrix_set (*q_matrix, i, j, gsl_matrix_get (A, i - 1, j - 1) );
        }
    }

    gsl_matrix_free (Id);
    gsl_matrix_free (A);
    gsl_matrix_free (A_transpose);
}

void Eigen_V(gsl_matrix * m, gsl_vector ** v, int n) {
    *v = gsl_vector_alloc (n);
    gsl_matrix * A = gsl_matrix_alloc (n, n);
    gsl_vector * eval = gsl_vector_alloc (n);
    gsl_matrix * evec = gsl_matrix_alloc (n, n);
    gsl_matrix_memcpy (A, m);
    gsl_eigen_symmv_workspace * w = gsl_eigen_symmv_alloc (n);
    gsl_eigen_symmv (A, eval, evec, w);
    gsl_eigen_symmv_free (w);
    int indice = 0;
    double max_v = fabs(gsl_vector_get(eval, 0));
    double cour;

    for (int i = 1; i < n; i++) {
        cour = fabs(gsl_vector_get(eval, i));

        if (cour > max_v) {
            max_v = cour;
            indice = i;
        }
    }

    gsl_vector_view ev = gsl_matrix_column (evec, indice);
    gsl_vector_memcpy (*v, &ev.vector);
    gsl_vector_free (eval);
    gsl_matrix_free (evec);
    gsl_matrix_free (A);
}

void Q_Rotation(gsl_vector * qr, gsl_matrix ** Rqr) {
    *Rqr = gsl_matrix_alloc (3, 3);
    gsl_matrix_set_zero (*Rqr);
    double cour;
    //Rqr[0][0]
    cour = gsl_vector_get(qr, 0) * gsl_vector_get(qr, 0) + gsl_vector_get(qr, 1) * gsl_vector_get(qr, 1) - gsl_vector_get(qr, 2) * gsl_vector_get(qr, 2) - gsl_vector_get(qr, 3) * gsl_vector_get(qr, 3);
    gsl_matrix_set (*Rqr, 0, 0, cour );
    //Rqr[1][1]
    cour = gsl_vector_get(qr, 0) * gsl_vector_get(qr, 0) + gsl_vector_get(qr, 2) * gsl_vector_get(qr, 2) - gsl_vector_get(qr, 1) * gsl_vector_get(qr, 1) - gsl_vector_get(qr, 3) * gsl_vector_get(qr, 3);
    gsl_matrix_set (*Rqr, 1, 1, cour );
    //Rqr[2][2]
    cour = gsl_vector_get(qr, 0) * gsl_vector_get(qr, 0) + gsl_vector_get(qr, 3) * gsl_vector_get(qr, 3) - gsl_vector_get(qr, 1) * gsl_vector_get(qr, 1) - gsl_vector_get(qr, 2) * gsl_vector_get(qr, 2);
    gsl_matrix_set (*Rqr, 2, 2, cour );
    //Rqr[1][0]
    cour = 2 * (gsl_vector_get(qr, 1) * gsl_vector_get(qr, 2) + gsl_vector_get(qr, 0) * gsl_vector_get(qr, 3));
    gsl_matrix_set (*Rqr, 1, 0, cour );
    //Rqr[0][1]
    cour = 2 * (gsl_vector_get(qr, 1) * gsl_vector_get(qr, 2) - gsl_vector_get(qr, 0) * gsl_vector_get(qr, 3));
    gsl_matrix_set (*Rqr, 0, 1, cour );
    //Rqr[2][0]
    cour = 2 * (gsl_vector_get(qr, 1) * gsl_vector_get(qr, 3) - gsl_vector_get(qr, 0) * gsl_vector_get(qr, 2));
    gsl_matrix_set (*Rqr, 2, 0, cour );
    //Rqr[0][2]
    cour = 2 * (gsl_vector_get(qr, 1) * gsl_vector_get(qr, 3) + gsl_vector_get(qr, 0) * gsl_vector_get(qr, 2));
    gsl_matrix_set (*Rqr, 0, 2, cour );
    //Rqr[2][1]
    cour = 2 * (gsl_vector_get(qr, 2) * gsl_vector_get(qr, 3) + gsl_vector_get(qr, 0) * gsl_vector_get(qr, 1));
    gsl_matrix_set (*Rqr, 2, 1, cour );
    //Rqr[1][2]
    cour = 2 * (gsl_vector_get(qr, 2) * gsl_vector_get(qr, 3) - gsl_vector_get(qr, 0) * gsl_vector_get(qr, 1));
    gsl_matrix_set (*Rqr, 1, 2, cour );
}

void Mem_Copy_Point_V(MyMesh::Point p, gsl_vector * u, int n) {
    for (int i = 0; i < 3; i++) {
        gsl_vector_set(u, i, p[i]);
    }

    if (n == 4) {
        gsl_vector_set(u, 3, 1);
    }
}

MyMesh::Point Access_Point(MyMesh *m, MyMesh::VertexHandle num_vertice) {
    MyMesh::Point p;
    p = (*m).point(num_vertice);
    return p;
}



void Mult_Matrice(gsl_vector * u, gsl_vector * v, gsl_matrix ** m) {
    *m = gsl_matrix_alloc (3, 3);
    double cour;
    gsl_matrix_set_zero (*m);

    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            cour = gsl_vector_get(u, i) * gsl_vector_get(v, j);
            gsl_matrix_set (*m, i, j, cour);
        }
    }
}

void Mult_Matrice_Vector(gsl_matrix * m, gsl_vector * v, int n) {
    gsl_vector * A = gsl_vector_alloc (n);
    gsl_vector_set_zero (A);
    double cour;

    for (int i = 0; i < n; i++) {
        cour = 0;

        for (int k = 0; k < n; k++) {
            cour += gsl_matrix_get (m, i, k) * gsl_vector_get(v, k);
        }

        gsl_vector_set (A, i, cour);
    }

    gsl_vector_memcpy (v, A);
    gsl_vector_free (A);
}

void Display_Matrix(gsl_matrix *m, int n1, int n2) {
    for (int i = 0; i < n1; i++) {
        for (int j = 0; j < n2; j++) {
            printf ("%g ", gsl_matrix_get (m, i, j));
        }

        printf("\n");
    }

    printf("\n");
}


void Display_Vector(gsl_vector *v, int n) {
    for (int i = 0; i < n; i++) {
        printf ("%g \n", gsl_vector_get (v, i));
    }

    printf("\n");
}

void Display_Transf (double** transformation) {
    for (int i = 0 ; i < 4 ; i++) {
        for (int j = 0 ; j < 4 ; j++) {
            printf("%lf ", transformation[i][j]);
        }

        printf("\n");
    }

    printf("\n");
}


double mean_square_error(MyMesh *source, MyMesh::VertexHandle *source_handle, MyMesh *model, MyMesh::VertexHandle *model_handle, double** transformation, int nb_vertices) {
    MyMesh::Point p_source;
    MyMesh::Point p_model;
    gsl_vector *p_vect = gsl_vector_alloc(4);
    gsl_matrix *m_trans = gsl_matrix_alloc(4, 4);
    gsl_matrix_set_zero(m_trans);

    for (int i = 0 ; i < 4 ; i++) {
        for (int j = 0 ; j < 4 ; j++) {
            gsl_matrix_set(m_trans, i, j, (double)transformation[i][j]);
        }
    }

    double sum = 0;

    for (int i = 0; i < nb_vertices; i++) {
        p_source = Access_Point(source, source_handle[i]);
        Mem_Copy_Point_V(p_source, p_vect, 4);
        Mult_Matrice_Vector(m_trans, p_vect, 4);
        p_model = Access_Point(model, model_handle[i]);
        sum +=  (p_model[0] - gsl_vector_get(p_vect, 0)) * (p_model[0] - gsl_vector_get(p_vect, 0)) +
                (p_model[1] - gsl_vector_get(p_vect, 1)) * (p_model[1] - gsl_vector_get(p_vect, 1)) +
                (p_model[2] - gsl_vector_get(p_vect, 2)) * (p_model[2] - gsl_vector_get(p_vect, 2));
    }

    sum /= (double)nb_vertices;
    gsl_vector_free(p_vect);
    gsl_matrix_free(m_trans);
    return sum;
}


void Copie_Mesh_Handle(MyMesh *m, MyMesh::VertexHandle ** m_handle, int nb_vertices) {
    (*m_handle) = (MyMesh::VertexHandle *)  malloc(nb_vertices * sizeof(MyMesh::VertexHandle));
    MyMesh::VertexIter v_it, v_end((*m).vertices_end());
    int i = 0;

    for (v_it = (*m).vertices_begin() ; v_it != v_end ; ++v_it) {
        (*m_handle)[i] = v_it.handle();
        i++;
    }
}

void Closest_Points (MyMesh *source, MyMesh *model, int nb_source, MyMesh::VertexHandle *** resultat) {
    (*resultat) = (MyMesh::VertexHandle **) malloc( 2 * sizeof(MyMesh::VertexHandle *) );

    for (int i = 0; i < 2; i++) {
        (*resultat)[i] = (MyMesh::VertexHandle* ) malloc( nb_source * sizeof(MyMesh::VertexHandle ) );
    }

    MyMesh::VertexIter source_it, source_end((*source).vertices_end());
    MyMesh::VertexIter model_it, model_end((*model).vertices_end());
    int index_source = 0;
    double distance = numeric_limits<double>::max();
    MyMesh::Point p1, p2;
    double buff;
    MyMesh::VertexHandle index, index_model;
    bool b;

    for (source_it = (*source).vertices_begin() ; source_it != source_end ; ++source_it) {
        p1 = (*source).point(source_it);
        b = false;

        for (model_it = (*model).vertices_begin() ; model_it != model_end ; ++model_it) {
            index_model = model_it.handle();

            if ( b == false) {
                index = index_model;
                b = true;
            }

            p2 = (*model).point(model_it);
            buff = sqrt((p1[0] - p2[0]) * (p1[0] - p2[0]) + (p1[1] - p2[1]) * (p1[1] - p2[1]) + (p1[2] - p2[2]) * (p1[2] - p2[2]));

            if (buff < distance) {
                distance = buff;
                index = index_model;
            }
        }

        (*resultat)[0][index_source] = source_it.handle();
        (*resultat)[1][index_source] = index;
        distance = numeric_limits<int>::max();
        index_source++;
    }
}

}
