#include "interface.h"

using namespace std;
using namespace icp2;
using namespace Preprocessing;
using namespace Correspondence;
using namespace Clusters;
using namespace chargementOFF;
using namespace Non_Rigid_Registration;

pMaillage p_maillage1, p_maillage2, p_maillage3;
int nb_vertices, nb_vertices2;
MyMesh source, model, source_ini;
MyMesh::VertexHandle ** resultat;
list<assignment> kernel;
list<Cluster*>* ls_clust = new list<Cluster*>;

MyMesh::Point *point1, *point2;

Command_line command_l;


void Viewer::loadModel(char* a, Command_line command) {
    command_l = command;

// read mesh from stdin
    if ( ! OpenMesh::IO::read_mesh(source, a) ) {
        std::cerr << "Error: Cannot read mesh from " << a << std::endl;
        return;
    }

    p_maillage1 = chargeMaillage(&source);

    if ( command_l.normals == 1 ) {
        point1 = new MyMesh::Point[source.n_vertices()];
        point2 = new MyMesh::Point[source.n_vertices()];
        gsl_vector *N = gsl_vector_calloc(3);

        for (int i = 0; i < source.n_vertices(); i++) {
            calculer_normales(&source, MyMesh::VertexHandle(i), &N);
            point1[i] = source.point(MyMesh::VertexHandle(i));
            point2[i][0] = (source.point(MyMesh::VertexHandle(i)))[0] + gsl_vector_get(N, 0);
            point2[i][1] = (source.point(MyMesh::VertexHandle(i)))[1] + gsl_vector_get(N, 1);
            point2[i][2] = (source.point(MyMesh::VertexHandle(i)))[2] + gsl_vector_get(N, 2);
        }
    }

    command_l.display_source = 1;
    command_l.display_model = 0;
    command_l.display_correspondence = 0;
    command_l.display_clusters = 0;
    command_l.display_deformation = 0;
}


void Viewer::loadModel(int argc, char** argv, Command_line command, int nb_off) {
    // read mesh from stdin
    if (nb_off >= 1) {
        if ( ! OpenMesh::IO::read_mesh(source, argv[1]) ) {
            std::cerr << "Error: Cannot read mesh from " << argv[1] << std::endl;
            return;
        }

        if (nb_off >= 2 && ! OpenMesh::IO::read_mesh(model, argv[2])) {
            std::cerr << "Error: Cannot read mesh from " << argv[2] << std::endl;
            return;
        }
    }

    double c0 = 0.7;
    double emax = 0.5;
    command_l = command;

    if ( command_l.mod_consistency == 1 ) {
        c0 = command_l.consistency;
    }

    if ( command_l.mod_emax == 1) {
        emax = command_l.Emax;
    }

    source_ini = MyMesh(source);
    resultat = NULL;
    nb_vertices = source.n_vertices();
    nb_vertices2 = model.n_vertices();
    Compute_Non_Rigid_Registration(&source_ini, &source, &model, &ls_clust, c0, emax, &resultat, command_l.N);
    p_maillage1 = chargeMaillage(&source_ini);
    p_maillage2 = chargeMaillage(&source);
    p_maillage3 = chargeMaillage(&model);
    // anciennes fonctions de chargement a partir d'un fichier OFF
    //p_maillage=chargeMaillage(f);
    //p_maillage2=chargeMaillage(f2);
}


void Viewer::draw() {
    float Red[9] = {1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0};
    float Green[9] = {0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0};
    float Blue[9] = {0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0};

    if ( command_l.display_source == 1) {
        traceMaillage(p_maillage1, Red);
    }

    if ( command_l.display_model == 1) {
        traceMaillage(p_maillage3, Green);
    }

    if ( command_l.display_correspondence == 1) {
        traceCorrespondence(&source_ini, &model, nb_vertices, resultat);
    }

    if ( command_l.display_clusters == 1) {
        traceClusters(p_maillage1, ls_clust);
    }

    if ( command_l.display_deformation == 1) {
        traceMaillage(p_maillage2, Blue);
    }

    if ( command_l.normals == 1 ) {
        for (int i = 0; i < source.n_vertices(); i++) {
            trace_V(point1[i], point2[i]);
        }
    }

    if ( command_l.fill_mesh == 1 ) {
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }
}

void Viewer::init() {
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    // Restore previous viewer state.
    restoreStateFromFile();
    // Opens help window
    help();
}

QString Viewer::helpString() const {
    QString text("<h2>Interface Non-Rigid Registration Under Isometric Deformations  </h2>");
    text += "Use the mouse to move the camera around the object. ";
    return text;
}




