#include "interface.h"
#include <qapplication.h>
#include "Non_Rigid_Registration.h"

using namespace std;
/* command line options */

bool is_off(const char *str);

void display_erreur() {
    cout << " ./interface [FILE] [OPTION] \n" << endl;
    cout << "[FILE] \n" << endl;
    cout << " source " << endl;
    cout << "     display only the mesh source to the screen. \n "  << endl;
    cout << " source model " << endl;
    cout << "     display the source and the model to the screen and compute the registration algorithm \n\n"  << endl;
    cout << " source model output " << endl;
    cout << "     display the source and the model to the screen, compute the registration algorithm and\n" << endl;
    cout << "     store the result in the mesh output \n"  << endl;
    cout << " Mesh files accepted : .off \n" << endl;
    cout << "[OPTION] \n" << endl;
    cout << " --normals  " << endl;
    cout << "     display the normals (only works for one input file ex: ./interface mymesh --normals)  \n"  << endl;
    cout << " --consistency c0  " << endl;
    cout << "     enter the threshold C0 used in the correspondence step "  << endl;
    cout << "     c0 is a value beetween 0 and 1. Default value C0=0.7 \n" << endl;
    cout << " --registration_error Emax  " << endl;
    cout << "     enter the threshold Emax used in the clustering step \n"  << endl;
    cout << "     Emax is a positive real value \n" << endl;
    cout << " --dont_display_source  " << endl;
    cout << "     do not display the first input mesh to the screen (ex: ./interface source model --display_source) \n"  << endl;
    cout << " --dont_display_model  " << endl;
    cout << "     do not display the second imput mesh to the screen (ex: ./interface source model --display_model) \n"  << endl;
    cout << " --display_correspondence  " << endl;
    cout << "     compute and display the correspondences beetween the source and the model to the screen \n" << endl;
    cout << " --display_deformation  " << endl;
    cout << "     compute and display the deformed mesh after the energy minimisation step \n" << endl;
    cout << " --display_clusters  " << endl;
    cout << "     compute and display the clusters beetween the source and the model to the screen \n"  << endl;
    cout << " --nb_iteration N " << endl;
    cout << "     enter the number N of iteration of the registration algorithm. Default Value N=1 \n"  << endl;
    cout << " --fill  " << endl;
    cout << "     fill all the meshes. Default: meshes are displayed with the LINES OpenGL mode \n"  << endl;
    cout << " --save  " << endl;
    cout << "     save steps of preprocessing and clustering and store data in files .sav  \n"  << endl;
    cout << "     this option is used only for N=1 (ex: ./interface mymesh1.off mymesh2.off save \n"  << endl;
    cout << " --saved  " << endl;
    cout << "     load data from files .sav only if a state has been saved before with the option save \n"  << endl;
    cout << "     this option is used only for N=1 (ex: ./interface mymesh1.off mymesh2.off saved \n"  << endl;
}


int main(int argc, char** argv) {
    cout << "Non-Rigid Registration Under Isometric Deformations \n" << endl;
// Read command lines arguments.
    QApplication application(argc, argv);
    // Instantiate the viewer.
    Viewer viewer;
    Command_line command;
    command.mod_consistency = 0;
    command.consistency = 0.7;
    command.mod_emax = 0;
    command.Emax = 0.5;
    command.display_source = 1;
    command.display_model = 1;
    command.display_correspondence = 0;
    command.display_clusters = 0;
    command.display_deformation = 0;
    command.fill_mesh = 0;
    command.normals = 0;
    command.N = 1;
    /*
      if (argc<=4){

      for (int i = 2; i < argc; i++) {

      if (!strcmp(argv[i], "--normals") ){
      command.normals = 1;
      }else if (!strcmp(argv[i], "--fill") ){
      command.fill_mesh = 1;
      }else{
      cout << "ERROR: Unknown option " << argv[i]  << endl;
      display_erreur();
      return 0;
      }
      }
      viewer.loadModel(argv[1],command);
      }else{*/
    int nb_off = 0;
    int i = 1;

    while ( i < argc && is_off(argv[i]) ) {
        nb_off++;
        i++;
    }

    if ( nb_off == 0) {
        display_erreur();
        return 0;
    } else if (nb_off == 1) {
        for (int i = nb_off + 1; i < argc; i++) {
            if (!strcmp(argv[i], "--normals") ) {
                command.normals = 1;
            } else if (!strcmp(argv[i], "--fill") ) {
                command.fill_mesh = 1;
            } else {
                cout << "ERROR: Unknown option " << argv[i]  << endl;
                display_erreur();
                return 0;
            }
        }

        viewer.loadModel(argv[1], command);
    } else {
        printf(" argc=%d \n", argc);

        for (int i = nb_off + 1; i < argc; i++) {
            if (!strcmp(argv[i], "--consistency") ) {
                command.mod_consistency = 1;
                command.consistency = (double) atof(argv[i+1]);
                i++;
            } else if (!strcmp(argv[i], "--registration_error") ) {
                command.mod_emax = 1;
                command.Emax = (double) atof(argv[i+1]);
                i++;
            } else if (!strcmp(argv[i], "--dont_display_source") ) {
                command.display_source = 0;
            } else if (!strcmp(argv[i], "--dont_display_model") ) {
                command.display_model = 0;
            } else if (!strcmp(argv[i], "--display_correspondence") ) {
                command.display_correspondence = 1;
            } else if (!strcmp(argv[i], "--display_clusters") ) {
                command.display_clusters = 1;
            } else if (!strcmp(argv[i], "--display_deformation") ) {
                command.display_deformation = 1;
            } else if (!strcmp(argv[i], "--nb_iteration") ) {
                command.N = atoi(argv[i+1]);;
                i++;
            } else if (!strcmp(argv[i], "--fill") ) {
                command.fill_mesh = 1;
            } else if (!strcmp(argv[i], "--save") ) {
                Non_Rigid_Registration::set_save();
            } else if (!strcmp(argv[i], "--saved") ) {
                Non_Rigid_Registration::set_saved();
            } else {
                cout << "ERROR: Unknown option " << argv[i]  << endl;
                display_erreur();
                return 0;
            }
        }

        viewer.loadModel(argc, argv, command, nb_off);
    }

#if QT_VERSION < 0x040000
    // Set the viewer as the application main widget.
    application.setMainWidget(&viewer);
#else
    viewer.setWindowTitle("Non-Rigid Registration Under Isometric Deformations");
#endif
    // Make the viewer window visible on screen.
    viewer.show();
    // Run main loop.
    return application.exec();
}


bool is_off(const char *str) {
    int size = 0;

    for (; str[size] != 0; ++size);

    if (str[size-4] != '.') return false;

    if (str[size-3] != 'o') return false;

    if (str[size-2] != 'f') return false;

    if (str[size-1] != 'f') return false;

    return true;
}
